    /*Project of game for amstad cpc 464.
    Copyright (C) 2013  Crappy Games(Antiono Lopez Quinto,David Mass Arroyo,Javier I. Cano Alberola,Jaume Gasa Gmez) 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.*/

#include <stdlib.h>
#include <stdio.h>

#include "actor.h"
#include "screen.h"
#include "time.h"
#include "sprite.h"
#include "animation.h"
#include "motionpath.h"

// =======================================================================
//	CreateMovement
// -----------------------------------------------------------------------
//	Creates an empty movement.
// =======================================================================
void CreateMovement(Movement* movement)
{
	movement->x 				= 0;
	movement->y 				= 0;
	movement->active 			= false;
	movement->repeat			= false;
	movement->animateOnStop 	= false;
	movement->direction		= DIR_DOWN;
}

// =======================================================================
//	CreateMovementCoords
// -----------------------------------------------------------------------
//	Creates a movement with the given parameters.
// =======================================================================
void CreateMovementCoords(Movement* movement, UI8 xStart, UI8 yStart, Direction direction, UI8 amount)
{
	switch(direction)
	{
		case DIR_DOWN :
			movement->x = xStart;
			movement->y = yStart + amount;
			break;
		case DIR_RIGHT :
			movement->x = xStart + amount;
			movement->y = yStart;
			break;
		case DIR_LEFT :
			movement->x = xStart - amount;
			movement->y = yStart;
			break;
		case DIR_UP :
			movement->x = xStart;
			movement->y = yStart - amount;
			break;
	}
}

// =======================================================================
//	CreateActor
// -----------------------------------------------------------------------
//	Creates a fully customised actor (without animation).
// -----------------------------------------------------------------------
//	NOTES
//	The actor's direction during creation time must be different from the
//	desired direction. As you can see the assignment is adding 1 to the
//	desired direction because the function that sets the direction
//	properly (called at the end) does nothing if the given direction is
//	the same than the previous assigned one (take a look at the first
//	branch of that function).
// =======================================================================
Actor* CreateActor(UI8 id, UI8 xTile, UI8 yTile, UI8 xVel, UI8 yVel, UI8 width, UI8 height,
	Direction direction, const I8* sprite, bool visible, ActorType type)
{
	Actor* actor = (Actor*) malloc(sizeof(Actor));
	
	// If we don't have enough memory return an empty actor
	if(actor == NULL)
	{
		printf("Not enough memory to create the actor %u", id);
		return NULL;
	}
	
	actor->id 			= id;
	actor->screenCoords.x = xTile * K_TILE_WIDTH;
	actor->screenCoords.y = yTile * K_TILE_HEIGHT;
	actor->tileCoords.x 	= xTile;
	actor->tileCoords.y 	= yTile;
	actor->velocity.x 	= xVel;
	actor->velocity.y 	= yVel;
	actor->width 		= width;
	actor->height 		= height;
	actor->direction 		= direction + 1;	
	actor->sprite 		= NULL;
	actor->spriteData		= sprite;
	actor->visible 		= visible;
	actor->timer 		= 0;
	actor->type 			= type;
	actor->animation 		= NULL;
	
	SetActorDirection(direction, actor);
	CreateMovement(&actor->movement);
	CreateMotionPath(&actor->motionPath, K_DEFAULT_MOTION_PATH_SIZE);
	
	actor->movement.direction = direction;
	
	return actor;
}

// =======================================================================
//	CreateAnimatedActor
// -----------------------------------------------------------------------
//	Creates a fully customised animated actor.
// =======================================================================
Actor* CreateAnimatedActor(UI8 id, UI8 xTile, UI8 yTile, UI8 xVel, UI8 yVel, UI8 width,
	UI8 height, Direction direction, const I8* sprite, bool visible, ActorType type,
	UI8 frameTime, UI8 maxFrames, bool playAnimation)
{
	Actor* actor = CreateActor(id, xTile, yTile, xVel, yVel, width, height, direction, sprite,
		visible, type);
	
	// If we don't have enough memory return an empty actor
	if(actor == NULL)
	{
		return NULL;
	}
	
	actor->animation = CreateAnimation(frameTime, maxFrames, width, height);
	
	// If we don't have enough memory return an actor without animation
	if(actor->animation == NULL)
	{
		printf("Not enough memory to create the actor's %u animation", id);
		return actor;
	}
	
	if(playAnimation)
		PlayAnimation(actor->animation);
	
	return actor;
}

// =======================================================================
//	DeleteActor
// -----------------------------------------------------------------------
//	Frees the memory used by the actor.
// =======================================================================
void DeleteActor(Actor* actor)
{
	if(actor->animation != NULL)
		DeleteAnimation(actor->animation);
	
	free(actor);
}

// =======================================================================
//	SetActorScreenPosition
// -----------------------------------------------------------------------
//	Sets the actor's x and y position in screen units (pixels).
// =======================================================================
void SetActorScreenPosition(UI8 x, UI8 y, Actor* actor)
{
	actor->tileCoords.x 	= x / K_TILE_WIDTH;
	actor->tileCoords.y 	= y / K_TILE_HEIGHT;
	actor->screenCoords.x = x;
	actor->screenCoords.y = y;
}

// =======================================================================
//	SetActorTilePosition
// -----------------------------------------------------------------------
//	Sets the actor's x and y position in tile units.
// =======================================================================
void SetActorTilePosition(UI8 x, UI8 y, Actor* actor)
{
	actor->tileCoords.x 	= x;
	actor->tileCoords.y 	= y;
	actor->screenCoords.x = x * K_TILE_WIDTH;
	actor->screenCoords.y = y * K_TILE_HEIGHT;
}

// =======================================================================
//	SetActorPosition
// -----------------------------------------------------------------------
//	Sets the direction the actor is facing and changes it's sprite
//	accordingly.
// =======================================================================
void SetActorDirection(Direction direction, Actor* actor)
{
	if(actor->direction == direction)
		return;
	
	actor->direction = direction;
	
	if(actor->animation != NULL)
	{
		actor->sprite = actor->spriteData + actor->width * actor->height * actor->animation->maxFrames * direction;
	}
	else
	{
		actor->sprite = actor->spriteData + actor->width * actor->height * direction;
	}
}

// =======================================================================
//	DrawActor
// -----------------------------------------------------------------------
//	Draws an actor in the screen.
// =======================================================================
void DrawActor(Actor* actor)
{
	if(!actor->visible)
		return;
	
	if(actor->animation != NULL && IsAnimationRunning(actor->animation))
	{
		actor->sprite = actor->spriteData + (actor->width * actor->height * actor->animation->maxFrames * actor->direction) + (actor->width * actor->height * actor->animation->frame);
	}
	
	DrawSprite(actor->screenCoords.x, actor->screenCoords.y, actor->width, actor->height, actor->sprite);
}

// =======================================================================
//	UpdateActor
// -----------------------------------------------------------------------
//	Updates the actor attributes.
// =======================================================================
void UpdateActor(Actor* actor)
{
	UI16 minTime = 0;
	
	if(actor->movement.active)
	{
		// ! NEEDS OPTIMIZATION
		switch(actor->direction)
		{
			case DIR_DOWN 	: minTime = K_TIME_UNITS_PER_SECOND / actor->velocity.y; break;
			case DIR_RIGHT	: minTime = K_TIME_UNITS_PER_SECOND / actor->velocity.x; break;
			case DIR_LEFT 	: minTime = K_TIME_UNITS_PER_SECOND / actor->velocity.x; break;
			case DIR_UP 		: minTime = K_TIME_UNITS_PER_SECOND / actor->velocity.y; break;
			default : break;
		}
		
		if(GetTime() - actor->timer >= minTime)
		{
			switch(actor->direction)
			{
				case DIR_DOWN 	: SetActorScreenPosition(actor->screenCoords.x, actor->screenCoords.y + 1, actor); break;
				case DIR_RIGHT	: SetActorScreenPosition(actor->screenCoords.x + 1, actor->screenCoords.y, actor); break;
				case DIR_LEFT 	: SetActorScreenPosition(actor->screenCoords.x - 1, actor->screenCoords.y, actor); break;
				case DIR_UP 		: SetActorScreenPosition(actor->screenCoords.x, actor->screenCoords.y - 1, actor); break;
				default : break;
			}
			
			actor->timer = GetTime();
			
			//UpdateActorMovement(actor);
		}
	}
	
	if(actor->animation != NULL)
		UpdateAnimation(actor->animation);
	/*
	UI16 minTime = 0;
	
	if(actor->moving)
	{
		// ! NEEDS OPTIMIZATION
		switch(actor->direction)
		{
			case DIR_DOWN 	: minTime = K_TIME_UNITS_PER_SECOND / actor->velocity.y; break;
			case DIR_RIGHT	: minTime = K_TIME_UNITS_PER_SECOND / actor->velocity.x; break;
			case DIR_LEFT 	: minTime = K_TIME_UNITS_PER_SECOND / actor->velocity.x; break;
			case DIR_UP 		: minTime = K_TIME_UNITS_PER_SECOND / actor->velocity.y; break;
			default : break;
		}
		
		if(GetTime() - actor->timer >= minTime)
		{
			switch(actor->direction)
			{
				case DIR_DOWN 	: SetActorScreenPosition(actor->screenCoords.x, actor->screenCoords.y + 1, actor); break;
				case DIR_RIGHT	: SetActorScreenPosition(actor->screenCoords.x + 1, actor->screenCoords.y, actor); break;
				case DIR_LEFT 	: SetActorScreenPosition(actor->screenCoords.x - 1, actor->screenCoords.y, actor); break;
				case DIR_UP 		: SetActorScreenPosition(actor->screenCoords.x, actor->screenCoords.y - 1, actor); break;
				default : break;
			}
			
			actor->timer = GetTime();
			
			//UpdateActorMotionPath(actor);
			//UpdateActorMovement(actor);
		}
	}
	
	if(actor->animation != NULL)
		UpdateAnimation(actor->animation);
	*/
}

// =======================================================================
//	PrintActor
// -----------------------------------------------------------------------
//	Prints the actor's data in the screen.
// =======================================================================
void PrintActor(UI8 x, UI8 y, Actor* actor, bool clearScreen)
{
	if(clearScreen)
		ClearScreen(K_DEFAULT_CLEAR_SCREEN_COLOR);
		
	SetCursor(x, y);
	
	printf("ActorAddr:\t0x%x\r\n", actor);
	printf("Id:\t%x\r\n", actor->id);
	printf("ScrCoords:\t(%u, %u)\r\n", actor->screenCoords.x, actor->screenCoords.y);
	printf("TileCoords:\t(%u, %u)\r\n", actor->tileCoords.x, actor->tileCoords.y);
	printf("Velocity:\t(%u, %u)\r\n", actor->velocity.x, actor->velocity.y);
	printf("MovActive:\t%u\r\n", actor->movement.active);
	printf("MotPathCount:\t%u\r\n", actor->motionPath.count);
	printf("Width:\t%u\r\n", actor->width);
	printf("Height:\t%u\r\n", actor->height);
	
	switch(actor->direction)
	{
		case DIR_DOWN	: printf("Dir:\tDOWN\r\n"); break;
		case DIR_RIGHT	: printf("Dir:\tRIGHT\r\n"); break;
		case DIR_LEFT	: printf("Dir:\tLEFT\r\n"); break;
		case DIR_UP		: printf("Dir:\tUP\r\n"); break;
		default : break;
	}
	
	printf("SpriteAddr:\t0x%x\r\n", actor->sprite);
	
	switch(actor->movement.active)
	{
		case true : printf("Moving:\tT\r\n"); break;
		case false : printf("Moving:\tF\r\n"); break;
		default : break;
	}
	
	switch(actor->visible)
	{
		case true : printf("Visible:\tT\r\n"); break;
		case false : printf("Visible:\tF\r\n"); break;
		default : break;
	}
	
	switch(actor->type)
	{
		case AT_PLAYER	: printf("Type:\tPLAYER\r\n"); break;
		case AT_ENEMY	: printf("Type:\tENEMY\r\n"); break;
		case AT_ITEM	: printf("Type:\tITEM\r\n"); break;
		default : break;
	}
	
	if(actor->animation != NULL)
		PrintAnimation(1, 15, actor->animation, false);
}
